#
# Copyright 2019- zhangpeng
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

require "fluent/plugin/output"
require 'tempfile'
require 'baidubce/services/bos/bos_client'


module Fluent
  module Plugin
    class BosOutput < Fluent::Plugin::Output
      include Baidubce
      Fluent::Plugin.register_output("bos", self)
      helpers :inject, :formatter, :compat_parameters, :thread
      DEFAULT_LINE_FORMAT_TYPE = 'single_value'
      DEFAULT_FORMAT_TYPE = 'json'

      desc "BCE access key id"
      config_param :bce_key_id, :string, secret: true
      desc "BCE secret key"
      config_param :bce_sec_key, :string, secret: true
      desc "bos endpoint"
      config_param :bos_endpoint, :string
      desc "bos bucket name"
      config_param :bos_bucket, :string
      desc "part name prefix"
      config_param :part_prefix, :string

      config_section :buffer do
        config_set_default :chunk_keys, ['tag']
        config_set_default :flush_at_shutdown, false
        config_set_default :chunk_limit_size, 10 * 1024
      end

      config_section :format do
        config_set_default :@type, DEFAULT_LINE_FORMAT_TYPE
        config_set_default :output_type, DEFAULT_FORMAT_TYPE
      end

      def prefer_buffered_processing
        true
      end

      def multi_workers_ready?
        true
      end

      attr_accessor :formatter

      def configure(conf)
        compat_parameters_convert(conf, :inject, :formatter) 
        super
  
        @formatter = formatter_create
      end

      def start 
        credentials = Auth::BceCredentials.new(
            @bce_key_id,
            @bce_sec_key
        )

        bosconf = BceClientConfiguration.new(
            credentials,
            @bos_endpoint
        )
        @client = Services::BosClient.new(bosconf)
        @part_suffix = 1
        @upload_id = @client.initiate_multipart_upload(@bos_bucket, @part_prefix+@part_suffix.to_s)["uploadId"]
        @part_number=1
        @part_list=[]
        @queue = Queue.new
        super
        thread_create(:send_tmpfile_to_bos, &method(:send_tmpfile))
      end
      
      def send_tmpfile
        while thread_current_running?
          if @part_number <= 100
              if @queue.size > 1
                  to_send = @queue.pop
                  tmpfile = to_send["file"]
                  tmpfile.rewind
                  s = tmpfile.read
                  tmpfile.close
                  tmpfile.unlink
                  response = @client.upload_part(
                      @bos_bucket, @part_prefix+@part_suffix.to_s, @upload_id, @part_number,s.size ,{})  do |buf_writer|
                             buf_writer << s
                  end
                  @part_list << {
                    "partNumber" => @part_number,
                    "eTag" => response['etag']
                  }
                  @part_number += 1 
              elsif @queue.size == 1
                   to_send = @current_file
                   if Time.now - to_send["time"] > 900
                     @queue.pop
                     tmpfile = to_send["file"]
                     tmpfile.rewind
                     s = tmpfile.read
                     tmpfile.close
                     tmpfile.unlink
                     response = @client.upload_part(
                         @bos_bucket, @part_prefix+@part_suffix.to_s, @upload_id, @part_number,s.size ,{})  do |buf_writer|
                             buf_writer << s
                     end
                     @part_list << {
                       "partNumber" => @part_number,
                        "eTag" => response['etag']
                     }
                    @client.complete_multipart_upload(@bos_bucket, @part_prefix+@part_suffix.to_s, @upload_id, @part_list)
                    @part_suffix += 1
                    @upload_id = @client.initiate_multipart_upload(@bos_bucket, @part_prefix+@part_suffix.to_s)["uploadId"]
                    @part_list = []
                    @part_number = 1
                   end
              end 
          elsif 
             @client.complete_multipart_upload(@bos_bucket, @part_prefix+@part_suffix.to_s, @upload_id, @part_list)
             @part_suffix += 1
             @upload_id = @client.initiate_multipart_upload(@bos_bucket, @part_prefix+@part_suffix.to_s)["uploadId"]
             @part_list = []
             @part_number = 1
          end 
        end
      end

      def format(tag, time, record)
        record = inject_values_to_record(tag, time, record)
        @formatter.format(tag, time, record).chomp + "\n"
      end

      def write(chunk)
          if @current_tmpfile.nil? || @current_tmpfile.path.nil? || @current_tmpfile.size >= 5*1024*1024
          @current_tmpfile = Tempfile.new("bostmpfile")
          @current_file = {
            "time" => Time.now,
            "file" => @current_tmpfile
          }
          @queue.push(@current_file)
        end
        ss = chunk.read
        @current_tmpfile.write(ss)
        @current_tmpfile.size
      end
    end
  end
end
